/* 
    bcc-lattice (c) 2000-2007 JS 

    v0.4 Oct07 new format
    v0.3 Nov00 output translation symmetries

*/
#include <stdlib.h>	/* Achtung bei long<->int-Mischmasch */
#include <stdio.h>

/* #define NN 128	*/	/* maximum value 8*8*8/4 */
#define NN 2000	/* maximum value 20*20*20/4 20121-12 */
/* 2012-12 4^3=9ms 884=0.8s 8^3=8.6s 88.16=136s 10^3=124s Rij=..slow t(n^4) */

/*   BCC-Gitter     */
#define Nw ((4+3)*NN)	/* J1-J2: ((4+3)*NN),  J1: ((4)*NN) */

int  pos[NN][3];			/* site-position = (x,y,z)	*/
int  iw[Nw][2];				/* array bonds		*/
int  wj[Nw];		 	/* nhv faster if wj4=wj/4	*/

#define F0(txt)	{ fprintf(stderr,"\nERROR L%d: "txt"\n",__LINE__);exit(1); }

/* ================== 3D-Gitter nach Betts =========================== */
/* 
   2 ineinandergeschachtelte cubische Gitter mit verdoppelter Einheitszelle
   das 2-te Gitter ist um (1,1,1) verschoben 
   -> utils/cubes.c
   Berechnung kubischer Gitter aus Vektoren die Elementarzelle aufspannen 
   e_cell = elementary cell (a,b,c) most compact form (-> Betts)
 */

int e_cell[3][3]={{4,0,0},{0,4,0},{0,0,4}};
/* think on modifying NN (first lines) */

int vertex[NN][3];              /* position of vertizees */         
int nn,nw;

/* 
+++TeX+++

 Beispiel in zwei Dimensionen. $a=(a_x,a_y)$ und $b=(b_x,b_y)$ sind die 
 zwei Vektoren, die die Elementarzelle aufspannen. 
 Sie sind im allgemeinen nicht senkrecht zueinander. 
 Der Vektor $r=(r_x,r_y)$ liegt innerhalb der Elementarzelle, wenn seine
 Komponenten $(r_a,r_b)$ bezueglich der Elementarvektoren zwischen $0$ und $1$
 liegen.
 \begin{eqnarray*}
      r &=& r_a a   + r_b b \\
 %
    a r &=& r_a a^2 + r_b a b \\
    b r &=& r_a a b + r_b b^2 \\
 %
    r_a &=& \frac{(b^2a-(a,b)b)r}{a^2b^2-(a,b)^2} \\
    r_b &=& \frac{(a^2b-(a,b)a)r}{a^2b^2-(a,b)^2} \\
 \end{eqnarray*}
 Grenzfall $(a,b)=0$ liefert bekannte Beziehungen. 
 Im 3 dimensionalen muss entsprechend ein Gleichungssystem 3.\ Ordnung 
 geloest werden.
 \begin{verbatim}
  maple> solve( { ar=ra*aa+rb*ab+rc*ac,\ 
  maple>          br=ra*ab+rb*bb+rc*bc,\
  maple>          cr=ra*ac+rb*bc+rc*cc},{ra,rb,rc});
  {ra= ( + (bb cc - bc bc) ar + (ac bc - cc ab) br + (bc ab - ac bb) cr)/%1,
   rb=-( + (ab cc - bc ac) ar + (ac ac - cc aa) br + (bc aa - ac ab) cr)/%1,
   rc= ( + (bc ab - bb ac) ar + (ab ac - aa bc) br + (aa bb - ab ab) cr)/%1}
  %1 := + ac bc ab + ac bc ab + bb cc aa - bc bc aa - ac ac bb - ab ab cc
  \end{verbatim}
  Auf die Division durch die Determinante kann evl. verzichtet werden, in dem
  die Einheit auf Det(a,b,c) gesetzt wird.
  Hilfreich koennte die Speicherung der Determinante der Einheitszelle und
  der inversen Elementarzelle (normiert mit der Determinante) sein.
---TeX--- */

/* vertex -> (x,y,z) */
int get_koord(const int i,int s[3]){
  if(i<0 || i>=nn) { printf("error: get_koord, i out of range\n"); return -1; }
  s[0]=vertex[i][0];
  s[1]=vertex[i][1];
  s[2]=vertex[i][2];
  return 0;}

int get_det(int e[3][3]){
 return( e[0][0]*(e[1][1]*e[2][2]-e[1][2]*e[2][1])     /* ax(by cz - bz cy) */ 
        +e[0][1]*(e[1][2]*e[2][0]-e[1][0]*e[2][2])     /* ay(bz ca - bx cz) */
        +e[0][2]*(e[1][0]*e[2][1]-e[1][1]*e[2][0]));   /* az(bx cb - by cx) */
}  
  
/* (x,y,z) -> vertex */  
int get_vertex(int xx[3],int dx,int dy,int dz){
  int i,x[3],(*e)[3]=e_cell,aa,ab,ac,bb,bc,cc,ar,br,cr,ra,rb,rc,
      det,(*s)[3]=vertex;
  x[0]=xx[0]+dx;x[1]=xx[1]+dy;x[2]=xx[2]+dz;
  aa=e[0][0]*e[0][0]+e[0][1]*e[0][1]+e[0][2]*e[0][2];
  ab=e[0][0]*e[1][0]+e[0][1]*e[1][1]+e[0][2]*e[1][2];
  ac=e[0][0]*e[2][0]+e[0][1]*e[2][1]+e[0][2]*e[2][2];
  bb=e[1][0]*e[1][0]+e[1][1]*e[1][1]+e[1][2]*e[1][2];
  bc=e[1][0]*e[2][0]+e[1][1]*e[2][1]+e[1][2]*e[2][2];
  cc=e[2][0]*e[2][0]+e[2][1]*e[2][1]+e[2][2]*e[2][2];
  det= + ac*bc*ab + ac*bc*ab + bb*cc*aa
       - bc*bc*aa - ac*ac*bb - ab*ab*cc;
  ar=e[0][0]*x[0]   +e[0][1]*x[1]   +e[0][2]*x[2];
  br=e[1][0]*x[0]   +e[1][1]*x[1]   +e[1][2]*x[2];
  cr=e[2][0]*x[0]   +e[2][1]*x[1]   +e[2][2]*x[2];
  ra= ( + (bb*cc - bc*bc)*ar + (ac*bc - cc*ab)*br + (bc*ab - ac*bb)*cr);
  rb=-( + (ab*cc - bc*ac)*ar + (ac*ac - cc*aa)*br + (bc*aa - ac*ab)*cr);
  rc= ( + (bc*ab - bb*ac)*ar + (ab*ac - aa*bc)*br + (aa*bb - ab*ab)*cr);
  if(ra<0)ra=-((-ra-1)/det)-1;else ra/=det;
  if(rb<0)rb=-((-rb-1)/det)-1;else rb/=det;
  if(rc<0)rc=-((-rc-1)/det)-1;else rc/=det;
  x[0]-=ra*e[0][0]+rb*e[1][0]+rc*e[2][0];
  x[1]-=ra*e[0][1]+rb*e[1][1]+rc*e[2][1];
  x[2]-=ra*e[0][2]+rb*e[1][2]+rc*e[2][2];
  for(i=0;i<nn;i++) 
    if(x[0]==s[i][0] && x[1]==s[i][1] && x[2]==s[i][2]) return i;
#if 1
  fprintf(stderr,
  "error get_vertex: %2d %2d %2d -(%2d %2d %2d)-> %2d %2d %2d not found\n",
    xx[0]+dx,xx[1]+dy,xx[2]+dz,ra,rb,rc,x[0],x[1],x[2]);
#endif
  return -1;}

int vertex_init(){ /* vertex_init */
   int i,j,k,r[3],ra,rb,rc,aa,ab,ac,bb,bc,cc,ar,br,cr,det,
     emin[3],emax[3],r0[3],
     (*s)[3]=vertex,
     (*e)[3]=e_cell /* ,*a=e[0],*b=e[1],*c=e[2] */ ;
   /* plaetze finden! */
   
   printf("# E=");
   for(i=0;i<3;i++)printf(" (%2d %2d %2d)",e[i][0],e[i][1],e[i][2]);
   i=( e[0][0]*(e[1][1]*e[2][2]-e[1][2]*e[2][1])     /* ax(by cz - bz cy) */ 
      +e[0][1]*(e[1][2]*e[2][0]-e[1][0]*e[2][2])     /* ay(bz ca - bx cz) */
      +e[0][2]*(e[1][0]*e[2][1]-e[1][1]*e[2][0]));   /* az(bx cb - by cx) */
   printf("\n# 2/8*Vol=2/8*Det(E)=%d/4=%d nn=%u",i,i/4,nn);
   if(abs(i)!=4*nn) F0("confusion in cluster size");
   
   /* einhuellenden Kubus ermitteln (alle eck-koords checken) */
   for(i=0;i<3;i++){ 
     emin[i]=emax[i]=e[0][i]+e[1][i]+e[2][i]; 
     if(emin[i]>0)emin[i]=0;
     if(emax[i]<0)emax[i]=0;
     for(j=0;j<3;j++){
       if(e[j][i]<emin[i])emin[i]=e[j][i];
       if(e[j][i]>emax[i])emax[i]=e[j][i];
       for(k=j+1;k<3;k++){
         if(e[j][i]+e[k][i]<emin[i])emin[i]=e[j][i]+e[k][i];
         if(e[j][i]+e[k][i]>emax[i])emax[i]=e[j][i]+e[k][i];
   }}}
   printf("# emin=(%d %d %d) emax=(%d %d %d)",
   emin[0],emin[1],emin[2],emax[0],emax[1],emax[2]);
   
   i=0;
   aa=e[0][0]*e[0][0]+e[0][1]*e[0][1]+e[0][2]*e[0][2];
   ab=e[0][0]*e[1][0]+e[0][1]*e[1][1]+e[0][2]*e[1][2];
   ac=e[0][0]*e[2][0]+e[0][1]*e[2][1]+e[0][2]*e[2][2];
   bb=e[1][0]*e[1][0]+e[1][1]*e[1][1]+e[1][2]*e[1][2];
   bc=e[1][0]*e[2][0]+e[1][1]*e[2][1]+e[1][2]*e[2][2];
   cc=e[2][0]*e[2][0]+e[2][1]*e[2][1]+e[2][2]*e[2][2];
   det= + ac*bc*ab + ac*bc*ab + bb*cc*aa - bc*bc*aa - ac*ac*bb - ab*ab*cc;

   /* sublattice A */
   for(r[0]=emin[0];r[0]<=emax[0];r[0]+=2)
   for(r[1]=emin[1];r[1]<=emax[1];r[1]+=2)  /* ineffektiv! */
   for(r[2]=emin[2];r[2]<=emax[2];r[2]+=2){
    ar=e[0][0]*r[0]   +e[0][1]*r[1]   +e[0][2]*r[2];
    br=e[1][0]*r[0]   +e[1][1]*r[1]   +e[1][2]*r[2];
    cr=e[2][0]*r[0]   +e[2][1]*r[1]   +e[2][2]*r[2];
    ra= ( + (bb*cc - bc*bc)*ar + (ac*bc - cc*ab)*br + (bc*ab - ac*bb)*cr);
    rb=-( + (ab*cc - bc*ac)*ar + (ac*ac - cc*aa)*br + (bc*aa - ac*ab)*cr);
    rc= ( + (bc*ab - bb*ac)*ar + (ab*ac - aa*bc)*br + (aa*bb - ab*ab)*cr);
    if(ra>=0 && rb>=0 && rc>=0 && ra<det && rb<det && rc<det){
      s[i][0]=r[0];
      s[i][1]=r[1];
      s[i][2]=r[2]; i++; }
   }
   
   /* sublattice B (body centered) */
   for(r[0]=emin[0]+1;r[0]<=emax[0];r[0]+=2)
   for(r[1]=emin[1]+1;r[1]<=emax[1];r[1]+=2)  /* ineffektiv! */
   for(r[2]=emin[2]+1;r[2]<=emax[2];r[2]+=2){
     ar=e[0][0]*r[0]   +e[0][1]*r[1]   +e[0][2]*r[2];
     br=e[1][0]*r[0]   +e[1][1]*r[1]   +e[1][2]*r[2];
     cr=e[2][0]*r[0]   +e[2][1]*r[1]   +e[2][2]*r[2];
     ra= ( + (bb*cc - bc*bc)*ar + (ac*bc - cc*ab)*br + (bc*ab - ac*bb)*cr);
     rb=-( + (ab*cc - bc*ac)*ar + (ac*ac - cc*aa)*br + (bc*aa - ac*ab)*cr);
     rc= ( + (bc*ab - bb*ac)*ar + (ab*ac - aa*bc)*br + (aa*bb - ab*ab)*cr);
     if(ra>=0 && rb>=0 && rc>=0 && ra<det && rb<det && rc<det){
       s[i][0]=r[0]; s[i][1]=r[1]; s[i][2]=r[2]; i++; }
   }
   if(i!=nn)fprintf(stderr," error vertex_init i!=N i=%d N=%d\n",i,nn);

   /* --- Platzvektoren ausgeben --- */
   for(i=0;i<nn;i++) 
     printf("%s %2d %2d %2d",((i)?",":"\n# Xij="),s[i][0],s[i][1],s[i][2]);
     
   /* --- better to move i=0 to Ri=( 0 0 0 ) !?, 1998-270 JS */
   r0[0]=s[0][0]; r0[1]=s[0][1]; r0[2]=s[0][2];
     
   /* --- kuerzesten Vektor r (PRB) bestimmen --- */
   /* ToDo: to slow! O(n^4) test 8x8x8 ++ */
   for(i=0;i<nn;i++){
     pos[i][0]=r[0]=s[i][0]-r0[0]; 
     pos[i][1]=r[1]=s[i][1]-r0[1]; 
     pos[i][2]=r[2]=s[i][2]-r0[2];
     for(ar=-nn;ar<nn;ar++)
     for(br=-nn;br<nn;br++)
     for(cr=-nn;cr<nn;cr++)     /* (2*nn)^3 steps */    
#define SQ(x) ((x)*(x))       /* besser mit C++ und Vektor! */
#define SW(x) (s[i][x]-r0[x]+ar*e[0][x]+br*e[1][x]+cr*e[2][x])
     if(SQ( r[0]) +SQ( r[1]) +SQ( r[2])
       >SQ(SW(0)) +SQ(SW(1)) +SQ(SW(2))){
         pos[i][0]=r[0]=SW(0);
         pos[i][1]=r[1]=SW(1);
         pos[i][2]=r[2]=SW(2); }
     printf("%s %+2d %+2d %+2d",((i)?",":"\n# Rij="),r[0],r[1],r[2]);
     fflush(stdout);
     /* output 3*3+1 chars for better looking SiSj table */
   }
   printf("\n");
   return 0;
}


  /*  i=site 0..nn, d=direction 0..2, s=steps ..,-1,0,1,.. rc=i+s*vec_d */
  /* besser site(i,dx,dy[,dz])  dy<0 geht nicht !!!!  */
  int site_move(int i,int dx,int dy,int dz,...){ int r[3];
    if(get_koord(i,r)) return -1;	/* i out of range */
    i=get_vertex(r,dx,dy,dz);
    return(i);}

  /* Abstandsdefinition (z.B.) fuer Untergittermagnetisierung */
  int site_dist(int i,int j){ /* Abstandsdefinition: |dx|+|dy|+|dz| */
     int ri[3],rj[3];
     get_koord(i,ri);
     get_koord(j,rj);
     return(abs(ri[0]-rj[0])+abs(ri[1]-rj[1])+abs(ri[2]-rj[2]));}

  int initvars(){
    int i; 
    vertex_init();
#define Bd(l,a,b) for(i=0;i<nn;i++){ iw[l][0]=a; iw[l][1]=b; }
    Bd(i     ,i,site_move(i,1, 1, 1));
    Bd(i+1*nn,i,site_move(i,1, 1,-1));
    Bd(i+2*nn,i,site_move(i,1,-1, 1));
    Bd(i+3*nn,i,site_move(i,1,-1,-1));
    if(nw>4*nn){
      Bd(i+4*nn,i,site_move(i,2, 0, 0));
      Bd(i+5*nn,i,site_move(i,0, 2, 0));
      Bd(i+6*nn,i,site_move(i,0, 0, 2));
    }
    for(i=0;i< nw;i++)wj[i]=(i<4*nn)?1:2;
    return 0;}
    
int topo_imperfection(){ /* max. 1 8 26=Ao+Kantlen+8 l=0,1,3,5,7 6*l*l+12*l+.. */
  int site[NN];
  int i,j,dist,nf,ns,nss;
  for(i=0;i<nn;i++) site[i]=-1;
  printf("# It= 1"); site[0]=0;
  for( nss=dist=1; dist<nn && nss<nn; dist++){ /* manhattan distance */
    nf=ns=0; /* num sites per given distance */
    for(i=0;i<nw;i++) for(j=0;j<2;j++){
      if( site[ iw[i][  j] ] == dist-1 ) nf++;
      if( site[ iw[i][  j] ] == dist-1
       && site[ iw[i][1-j] ] <  0      ){ /* next neighbour */
        site[ iw[i][1-j] ] = dist;
        ns++; nss++;
      }
    }
    printf(" %d",ns);
  }
  printf(" (1 8 26 56(?) ...)\n");
  return 0;
}

int outmodel(){
  int i;
  /* Ausgabe Modell in Maschinen-lesbarer Form (fuer allg. Model.H) */
  printf("#+++model+++\n");
  printf("# NN=Number_of_sites Nw=Number_of_bonds Nv=Number_of_bonds(2)\n");
  printf("%2u %3u 0\n",nn,nw);
  printf("# s  X  Y  Z e# U# (s=site  (X Y Z)=position  eU=parameter-Nr)\n");
  for(i=0;i<nn;i++)
  printf(" %2u %2d %2d %2d e3 u4\n",i,pos[i][0],pos[i][1],pos[i][2]);
  printf("# Bond s1 s2 t# J# V#\n");
  for(i=0;i<nw;i++)
  printf(" %3u %2d %2d  t2 j%d v0\n",i,iw[i][0],iw[i][1],wj[i]);
  printf("#---model---\n");
  return(0); }

/* output translation as permutation of sites
 *  to use k-vectors
 */
int  outtranslation( int dx, int dy, int dz){
  int i,j;
  printf("# sym (%d,%d,%d):",dx,dy,dz);
  for (i=0;i<nn;i++){
     printf(" %d",get_vertex(vertex[i],dx,dy,dz));
  }
  printf("\n");
  return 0;
}

int main(int argn,char *argv[]){
  int i;
  printf("# m_bcc v0.1 Feb2000 JS\n");
  if(argn!=11){
    fprintf(stderr,
    "usage:   m_bcc l11 l12 l13 l21 l22 l23 l31 l32 l33 bounds\n"
    "example: m_bcc 4 0 0  0 4 0  0 0 4  2\n"
    "   edgevector=((l11,l12,l13),(l21,l22,l23),(l31,l32,l33))\n"
    "   bounds: 1=J1 2=J1+J2\n"
    );
    return 1;
  }
  for(i=0;i<9;i++)e_cell[i/3][i%3]=atoi(argv[i+1]);
  nn=get_det(e_cell)/4;
  if( nn>NN ) { fprintf(stderr,"nn=%d > NN=%d\n",nn,NN); exit(1); }
  if(atoi(argv[10])==2)
    nw=(4+3)*nn;		/* J1+J2 */
  else
    nw=4*nn;			/* J1 */
  initvars();
  topo_imperfection();
  outmodel();
  outtranslation( +2, +0, +0); /* (dx,dy,dz) */
  outtranslation( +0, +2, +0); /* (dx,dy,dz) */
  outtranslation( +0, +0, +2); /* (dx,dy,dz) */
  return 0;
}
